#This file is part of Tryton.  The COPYRIGHT file at the top level of
#this repository contains the full copyright notices and license terms.
from trytond.model import ModelWorkflow, ModelView, ModelSQL, fields
import datetime
from decimal import Decimal


class ProfitCenter(ModelSQL, ModelView):
    'Sale Import Profit Center'
    _name = 'sale_import.profit_center'
    _description = __doc__
    name = fields.Char('Name', required=True)
    storage_location = fields.Many2One('stock.location', 'Storage Location',
            domain=[('type', '=', 'storage')], required=True)
    sale_journal = fields.Many2One('account.journal', 'Sale Journal',
            required=True, domain=[('centralised', '=', False)])

ProfitCenter()


class Journal(ModelWorkflow, ModelSQL, ModelView):
    'Sale Import Journal'
    _name = 'sale_import.journal'
    _description = __doc__
    profit_center = fields.Many2One('sale_import.profit_center',
            'Profit Center', required=True, states={
                'readonly': "state != 'draft'",
                })
    date = fields.Date('Date', required=True, states={
        'readonly': "state != 'draft'",
        })
    state = fields.Selection([
        ('draft', 'Draft'),
        ('confirmed', 'Confirmed'),
        ], 'Status', required=True, readonly=True)
    lines = fields.One2Many('sale_import.journal.line', 'journal',
            'Lines', states={
                'readonly': "state != 'draft'",
                })
    company = fields.Many2One('company.company', 'Company', required=True,
            states={
                'readonly': "state != 'draft' or bool(lines)",
                })
    currency = fields.Function('get_currency', type='many2one',
            relation='currency.currency', string='Currency')
    currency_digits = fields.Function('get_currency_digits', type='integer',
            string='Currency Digits')
    total_amount = fields.Function('get_total_amount', type='numeric',
            digits="(16, currency_digits)", string="Total",
            on_change_with=['lines', 'company'])
    move = fields.Many2One('account.move', 'Move', readonly=True)

    def __init__(self):
        super(Journal, self).__init__()
        self._order.insert(0, ('date', 'DESC'))

    def default_date(self, cursor, user, context=None):
        date_obj = self.pool.get('ir.date')
        return date_obj.today(cursor, user, context=context)

    def default_state(self, cursor, user, context=None):
        return 'draft'

    def default_company(self, cursor, user, context=None):
        company_obj = self.pool.get('company.company')
        if context is None:
            context = {}
        if context.get('company'):
            return context['company']
        return False

    def default_currency(self, cursor, user, context=None):
        company_obj = self.pool.get('company.company')
        currency_obj = self.pool.get('currency.currency')
        if context is None:
            context = {}
        company = None
        if context.get('company'):
            company = company_obj.browse(cursor, user, context['company'],
                    context=context)
            return company.currency.id
        return False

    def default_currency_digits(self, cursor, user, context=None):
        company_obj = self.pool.get('company.company')
        if context is None:
            context = {}
        company = None
        if context.get('company'):
            company = company_obj.browse(cursor, user, context['company'],
                    context=context)
            return company.currency.digits
        return 2

    def on_change_with_total_amount(self, cursor, user, ids, vals,
            context=None):
        currency_obj = self.pool.get('currency.currency')
        company_obj = self.pool.get('company.company')

        res = Decimal('0.0')

        currency = None
        if vals.get('company'):
            currency = company_obj.browse(cursor, user, vals['company'],
                    context=context).currency
        else:
            currency_id = self.default_currency(cursor, user, context=context)
            if currency_id:
                currency_id = currency_id[0]
                currency = currency_obj.browse(cursor, user, currency_id,
                        context=context)

        if vals.get('lines'):
            for line in vals['lines']:
                res += line.get('amount', Decimal('0.0'))
        if currency:
            res = currency_obj.round(cursor, user, currency, res)
        return res

    def get_currency(self, cursor, user, ids, name, arg, context=None):
        res = {}
        journals = self.browse(cursor, user, ids, context=context)
        for journal in journals:
            res[journal.id] = journal.company.currency.id
        return res

    def get_currency_digits(self, cursor, user, ids, name, arg, context=None):
        res = {}
        for journal in self.browse(cursor, user, ids, context=context):
            res[journal.id] = journal.company.currency.digits
        return res

    def get_total_amount(self, cursor, user, ids, name, arg, context=None):
        currency_obj = self.pool.get('currency.currency')
        res = {}
        for journal in self.browse(cursor, user, ids, context=context):
            res.setdefault(journal.id, Decimal('0.0'))
            for line in journal.lines:
                res[journal.id] += line.amount
            res[journal.id] = currency_obj.round(cursor, user, journal.currency,
                    res[journal.id])
        return res

    def get_rec_name(self, cursor, user, ids, name, arg, context=None):
        lang_obj = self.pool.get('ir.lang')

        if context is None:
            context = {}

        if not ids:
            return []
        res = {}

        for code in [context.get('language', False) or 'en_US', 'en_US']:
            lang_ids = lang_obj.search(cursor, user, [
                ('code', '=', code),
                ], context=context)
            if lang_ids:
                break
        lang = lang_obj.browse(cursor, user, lang_ids[0], context=context)

        for journal in self.browse(cursor, user, ids, context=context):
            res[journal.id] = unicode(journal.date.strftime(str(lang.date)))
        return res

    def search_rec_name(self, cursor, user, name, args, context=None):
        args2 = []
        i = 0
        while i < len(args):
            args2.append(('date', args[i][1], args[i][2]))
            i += 1
        return args2

    def create_stock_move(self, cursor, user, journal_id, context=None):
        '''
        Create stock moves for each lines

        :param cursor: the database cursor
        :param user: the user id
        :param journal_id: the id of the journal
        :param context: the context
        '''
        line_obj = self.pool.get('sale_import.journal.line')

        journal = self.browse(cursor, user, journal_id, context=context)
        for line in journal.lines:
            line_obj.create_stock_move(cursor, user, line, context=context)

    def compute_amount_service(self, cursor, user, journal_id, context=None):
        '''
        Compute service amount for each lines

        :param cursor: the database cursor
        :param user: the user id
        :param journal_id: the id of the journal
        :param context: the context
        '''
        line_obj = self.pool.get('sale_import.journal.line')

        journal = self.browse(cursor, user, journal_id, context=context)
        for line in journal.lines:
            line_obj.compute_amount_service(cursor, user, line, context=context)

    def _get_account_move_line(self, cursor, user, journal, context=None):
        line_obj = self.pool.get('sale_import.journal.line')
        res = []
        for line in journal.lines:
            vals = line_obj.get_account_move_line(cursor, user, line,
                    context=context)
            res += vals
        return res

    def create_account_move(self, cursor, user, journal_id, context=None):
        '''
        Create account moves for each lines

        :param cursor: the database cursor
        :param user: the user id
        :param journal_id: the id of the journal
        :param context: the context
        '''
        move_obj = self.pool.get('account.move')
        move_line_obj = self.pool.get('account.move.line')
        period_obj = self.pool.get('account.period')
        property_obj = self.pool.get('ir.property')
        model_field_obj = self.pool.get('ir.model.field')

        journal = self.browse(cursor, user, journal_id, context=context)

        move_lines = self._get_account_move_line(cursor, user, journal,
                context=context)

        field_id = model_field_obj.search(cursor, user, [
            ('name', '=', 'account_receivable'),
            ('model.model', '=', 'party.party'),
            ], limit=1, context=context)[0]
        property_id = property_obj.search(cursor, user, [
            ('field', '=', field_id),
            ('res', '=', False),
            ], limit=1, context=context)
        if not property_id:
            self.raise_user_error(cursor, 'missing_account_receivable',
                    context=context)
        value = property_obj.browse(cursor, user, property_id[0],
                context=context).value
        if not value:
            self.raise_user_error(cursor, 'missing_account_receivable',
                    context=context)
        account_receivable_id = int(value.split(',')[1])

        period_id = period_obj.find(cursor, user, journal.company.id,
                date=journal.date, context=context)

        move_id = move_obj.create(cursor, user, {
            'journal': journal.profit_center.sale_journal.id,
            'period': period_id,
            'date': journal.date,
            'lines': [('create', x) for x in move_lines],
            }, context=context)
        self.write(cursor, user, journal.id, {
            'move': move_id,
            }, context=context)

        move_line_id = move_line_obj.create(cursor, user, {
            'name': str(journal.date),
            'debit': journal.total_amount,
            'credit': Decimal('0.0'),
            'account': account_receivable_id,
            'move': move_id,
            }, context=context)

        move_obj.post(cursor, user, move_id, context=context)

    def copy(self, cursor, user, ids, default=None, context=None):
        if default is None:
            default = {}
        default = default.copy()
        default['move'] = False
        return super(Journal, self).copy(cursor, user, ids,
                default=default, context=context)

Journal()


class JournalLine(ModelSQL, ModelView):
    'Sale Import Journal Line'
    _name = 'sale_import.journal.line'
    _description = __doc__
    journal = fields.Many2One('sale_import.journal', 'Journal', required=True)
    date = fields.DateTime('Date', required=True)
    quantity = fields.Float('Quantity', digits="(16, unit_digits)", required=True)
    unit = fields.Many2One('product.uom', 'Unit', required=True,
            domain="[('category', '=', " \
                    "(product, 'product.default_uom.category'))]",
            context="{'category': (product, 'product.default_uom.category')}")
    unit_digits = fields.Function('get_unit_digits', type='integer',
            string='Unit Digits', on_change_with=['unit'])
    product = fields.Many2One('product.product', 'Product', required=True)
    untaxed_amount = fields.Numeric('Untaxed Amount',
            digits="(16, _parent_journal.currency_digits)")
    amount = fields.Numeric('Amount',
            digits="(16, _parent_journal.currency_digits)")
    moves = fields.One2Many('stock.move', 'sale_import_journal_line', 'Moves',
            readonly=True)
    amount_service = fields.Numeric('Service Amount',
            digits="(16, 4)")

    def __init__(self):
        super(JournalLine, self).__init__()
        self._error_messages.update({
            'missing_customer_location': 'There is no default customer ' \
                    'location defined!',
            })

    def default_date(self, cursor, user, context=None):
        return datetime.datetime.now()

    def on_change_with_unit_digits(self, cursor, user, ids, vals,
            context=None):
        uom_obj = self.pool.get('product.uom')
        if vals.get('unit'):
            uom = uom_obj.browse(cursor, user, vals['unit'],
                    context=context)
            return uom.digits
        return 2

    def get_unit_digits(self, cursor, user, ids, name, arg, context=None):
        res = {}
        for line in self.browse(cursor, user, ids, context=context):
            if line.unit:
                res[line.id] = line.unit.digits
            else:
                res[line.id] = 2
        return res

    def get_rec_name(self, cursor, user, ids, name, arg, context=None):
        lang_obj = self.pool.get('ir.lang')

        if context is None:
            context = {}

        if not ids:
            return []
        res = {}

        for code in [context.get('language', False) or 'en_US', 'en_US']:
            lang_ids = lang_obj.search(cursor, user, [
                ('code', '=', code),
                ], context=context)
            if lang_ids:
                break
        lang = lang_obj.browse(cursor, user, lang_ids[0], context=context)

        for line in self.browse(cursor, user, ids, context=context):
            res[line.id] = unicode(line.date.strftime(str(lang.date)))
        return res

    def search_rec_name(self, cursor, user, name, args, context=None):
        args2 = []
        i = 0
        while i < len(args):
            args2.append(('date', args[i][1], args[i][2]))
            i + 1
        return args2

    def copy(self, cursor, user, ids, default=None, context=None):
        if default is None:
            default = {}
        default = default.copy()
        default['state'] = 'draft'
        default['moves'] = False
        return super(JournalLine, self).copy(cursor, user, ids,
                default=default, context=context)

    def _create_composite_stock_move(self, cursor, user, line,
            customer_location_id, product, quantity, uom, unit_price,
            context=None):
        uom_obj = self.pool.get('product.uom')

        for composite in product.composites:
            if composite.product.type == 'service':
                continue

            composite_quantity = uom_obj.compute_qty(cursor, user, uom,
                    quantity, product.default_uom, context=context) \
                            * composite.quantity
            composite_unit_price = unit_price * \
                    (Decimal(str(composite.quantity)) * \
                    composite.product.cost_price) / \
                    product.cost_price

            if not composite.product.composite:
                vals = {}
                vals['quantity'] = composite_quantity
                vals['uom'] = composite.product.default_uom
                vals['product'] = composite.product.id
                vals['from_location'] = line.journal.profit_center.\
                        storage_location.id
                vals['to_location'] = customer_location_id
                vals['state'] = 'done'
                vals['company'] = line.journal.company.id
                vals['unit_price'] = composite_unit_price
                vals['currency'] = line.journal.currency.id
                vals['planned_date'] = line.date.date()
                vals['effective_date'] = line.date.date()
                self.write(cursor, user, line.id, {
                    'moves': [('create', vals)],
                    }, context=context)
            else:
                self._create_composite_stock_move(cursor, user, line,
                        customer_location_id, composite.product,
                        composite_quantity, composite.product.default_uom,
                        composite_unit_price, context=context)


    def create_stock_move(self, cursor, user, line, context=None):
        '''
        Create stock move for the line

        :param cursor: the database cursor
        :param user: the user id
        :param line: a BrowseRecord of the journal line
        :param context: the context
        '''
        property_obj = self.pool.get('ir.property')
        model_field_obj = self.pool.get('ir.model.field')
        tax_obj = self.pool.get('account.tax')
        uom_obj = self.pool.get('product.uom')

        if line.product.type == 'service':
            return

        field_id = model_field_obj.search(cursor, user, [
            ('name', '=', 'customer_location'),
            ('model.model', '=', 'party.party'),
            ], limit=1, context=context)[0]
        property_id = property_obj.search(cursor, user, [
            ('field', '=', field_id),
            ('res', '=', False),
            ], limit=1, context=context)
        if not property_id:
            self.raise_user_error(cursor, 'missing_customer_location',
                    context=context)
        value = property_obj.browse(cursor, user, property_id[0],
                context=context).value
        if not value:
            self.raise_user_error(cursor, 'missing_customer_location',
                    context=context)
        customer_location_id = int(value.split(',')[1])

        price_unit = line.untaxed_amount / Decimal(str(line.quantity))

        if not line.product.composite:
            vals = {}
            vals['quantity'] = line.quantity
            vals['uom'] = line.unit.id
            vals['product'] = line.product.id
            vals['from_location'] = line.journal.profit_center.storage_location.id
            vals['to_location'] = customer_location_id
            vals['state'] = 'done'
            vals['company'] = line.journal.company.id
            vals['unit_price'] = price_unit
            vals['currency'] = line.journal.currency.id
            vals['planned_date'] = line.date.date()
            vals['effective_date'] = line.date.date()
            self.write(cursor, user, line.id, {
                'moves': [('create', vals)],
                }, context=context)
        else:
            unit_price = uom_obj.compute_price(cursor, user, line.unit,
                    price_unit, line.product.default_uom,
                    context=context)
            self._create_composite_stock_move(cursor, user, line,
                    customer_location_id, line.product, line.quantity,
                    line.unit, unit_price, context=context)

    def compute_amount_service(self, cursor, user, line, context=None):
        '''
        Create service amount for the line

        :param cursor: the database cursor
        :param user: the user id
        :param line: a BrowseRecord of the journal line
        :param context: the context
        '''
        uom_obj = self.pool.get('product.uom')

        if not line.product.composite:
            return

        amount_service = Decimal('0')
        for composite in line.product.composites:
            if composite.product.type != 'service':
                continue
            if composite.profit_center and \
                    composite.profit_center.id != line.journal.profit_center.id:
                continue

            comp_cost_price = uom_obj.compute_price(
                cursor, user, line.product.default_uom, composite.product.cost_price,
                line.unit, context=context)

            amount_service += comp_cost_price *  Decimal(str(composite.quantity)) *  \
                Decimal(str(line.quantity))

        self.write(
            cursor, user, line.id, {'amount_service': amount_service},
            context=context)

    def get_account_move_line(self, cursor, user, line, context=None):
        '''
        Return the values for the move line of the line

        :param cursor: the database cursor
        :param user: the user id
        :param line: a BrowseRecord of the journal line
        :param context: the context

        :return: a dictionary of values of move line
        '''
        tax_obj = self.pool.get('account.tax')
        currency_obj = self.pool.get('currency.currency')

        res = []

        price_unit = line.untaxed_amount / Decimal(str(line.quantity))
        tax_ids = [x.id for x in line.product.customer_taxes_used]

        taxes = tax_obj.compute(cursor, user, tax_ids, price_unit,
                line.quantity, context=context)

        tax_amount = Decimal('0.0')
        for tax in taxes:
            tax_amount += tax['amount']

        diff = line.amount - line.untaxed_amount - tax_amount

        for tax in taxes:
            if tax['amount'] != Decimal('0.0'):
                tax['amount'] += diff * \
                        (tax['amount'] / tax_amount)
                tax['amount'] = currency_obj.round(cursor, user,
                        line.journal.currency, tax['amount'])

        # Re-check amount to add rounding error to first tax
        tax_amount = Decimal('0.0')
        for tax in taxes:
            tax_amount += tax['amount']

        diff = line.amount - line.untaxed_amount - tax_amount
        if not currency_obj.is_zero(cursor, user, line.journal.currency,
                diff):
            for tax in taxes:
                if tax['amount'] != Decimal('0.0'):
                    tax['amount'] += diff
                    break

        credit_amout = line.untaxed_amount
        if diff != Decimal('0.0') and not taxes:
            credit_amout = line.amount

        val = {}
        val['name'] = line.product.name
        val['debit'] = Decimal('0.0')
        val['credit'] = currency_obj.round(cursor, user, line.journal.currency,
                credit_amout)
        val['account'] = line.product.account_revenue_used.id

        for tax in taxes:
            if tax['tax'].invoice_base_code.id:
                val.setdefault('tax_lines', [])
                res['tax_lines'].append(('create', {
                    'code': tax['tax'].invoice_base_code.id,
                    'amount': currency_obj.round(cursor, user,
                        line.journal.currency,
                        tax['base'] * tax['tax'].invoice_base_sign),
                    }))
        res.append(val)

        for tax in taxes:
            val = {}
            val['name'] = tax['tax'].description
            if tax['amount'] >= Decimal('0.0'):
                val['debit'] = Decimal('0.0')
                val['credit'] = currency_obj.round(cursor, user,
                        line.journal.currency, tax['amount'])
            else:
                val['debit'] = currency_obj.round(cursor, user,
                        line.journal.currency, - tax['amount'])
                val['credit'] = Decimal('0.0')
            val['account'] = tax['tax'].invoice_account.id
            if tax['tax'].invoice_tax_code.id:
                val['tax_lines'] = [('create', {
                    'code': tax['tax'].invoice_tax_code.id,
                    'amount': currency_obj.round(cursor, user,
                        line.journal.currency,
                        tax['amount'] * tax['tax'].invoice_tax_sign),
                    })]
            res.append(val)
        return res

JournalLine()


class Move(ModelSQL, ModelView):
    _name = 'stock.move'

    sale_import_journal_line = fields.Many2One('sale_import.journal.line',
            readonly=True)

Move()
